/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.tika.parser.microsoft.ooxml;

import org.apache.poi.openxml4j.exceptions.OpenXML4JException;
import org.apache.poi.openxml4j.opc.*;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.xssf.eventusermodel.*;
import org.apache.poi.xssf.extractor.XSSFEventSimpleExcelExtractor;
import org.apache.poi.xssf.model.StylesTable;
import org.apache.poi.xssf.usermodel.*;
import org.apache.tika.exception.TikaException;
import org.apache.tika.metadata.Metadata;
import org.apache.tika.metadata.TikaMetadataKeys;
import org.apache.tika.sax.XHTMLContentHandler;
import org.apache.xmlbeans.XmlException;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

public class XSSFExcelSimpleExtractorDecorator extends AbstractOOXMLExtractor {

    /**
     * Internal <code>DataFormatter</code> for formatting Numbers.
     */
    private final DataFormatter formatter;

    private final XSSFEventSimpleExcelExtractor extractor;
    private static final String TYPE = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
    private boolean protectedWorkbook = false;

    public XSSFExcelSimpleExtractorDecorator(
            XSSFEventSimpleExcelExtractor extractor, Locale locale) {
        super(extractor, TYPE);

        this.extractor = extractor;                                        
        formatter = new DataFormatter(locale);
    }

    /**
     * @see org.apache.poi.xssf.extractor.XSSFExcelExtractor#getText()
     */
    @Override
    protected void buildXHTML(XHTMLContentHandler xhtml) throws SAXException,
            XmlException, IOException {
        XSSFSimpleWorkbook document = (XSSFSimpleWorkbook) extractor.getDocument();
        try{
            ReadOnlySharedStringsTable strings = new ReadOnlySharedStringsTable(document.getContainerPkg());
            XSSFReader xssfReader = new XSSFReader(document.getContainerPkg());
            StylesTable styles = xssfReader.getStylesTable();
            XSSFReader.SheetIterator iter = (XSSFReader.SheetIterator) xssfReader.getSheetsData();
            while (iter.hasNext()) {
                InputStream stream = iter.next();
                xhtml.startElement("div");
                xhtml.element("h1", iter.getSheetName());
                processSheet(styles, strings, stream, xhtml);
                stream.close();
            }
        } catch(IOException e) {
        System.err.println(e);
        } catch(SAXException se) {
        System.err.println(se);
        } catch(OpenXML4JException o4je) {
        System.err.println(o4je);
        }
    }

    /**
    * Processes the given sheet
    */
   protected void processSheet(
           StylesTable styles,
           ReadOnlySharedStringsTable strings,
           InputStream sheetInputStream,
           XHTMLContentHandler xhtml)
           throws IOException, SAXException {

       InputSource sheetSource = new InputSource(sheetInputStream);
       SAXParserFactory saxFactory = SAXParserFactory.newInstance();
       try {
          SAXParser saxParser = saxFactory.newSAXParser();
          XMLReader sheetParser = saxParser.getXMLReader();
          EventBasedXSSFSheetHandler handler = new EventBasedXSSFSheetHandler(styles, strings, xhtml);
          sheetParser.setContentHandler(handler);
          sheetParser.parse(sheetSource);
           if(handler.isSheetProtected()){
                this.protectedWorkbook = true;
           }           
       } catch(ParserConfigurationException e) {
          throw new RuntimeException("SAX parser appears to be broken - " + e.getMessage());
       }
   }

    /**
     * In Excel files, sheets have things embedded in them,
     *  and sheet drawings which have the images
     */
    @Override
    protected List<PackagePart> getMainDocumentParts() throws TikaException {
       List<PackagePart> parts = new ArrayList<PackagePart>();
//TODO checking how this works with SAXparsing
/*       XSSFSimpleWorkbook document = (XSSFSimpleWorkbook) extractor.getDocument();
       try {
            Iterator<XSSFSimpleSheet> it = document.getSheets().iterator();
            while(it.hasNext()){
                XSSFSimpleSheet sheet = it.next();
                PackagePart part = sheet.getPackagePart();

              // Add the sheet
                parts.add(part);

              // If it has drawings, return those too
                try {
                 for(PackageRelationship rel : part.getRelationshipsByType(XSSFRelation.DRAWINGS.getRelation())) {
                    if(rel.getTargetMode() == TargetMode.INTERNAL) {
                       PackagePartName relName = PackagingURIHelper.createPartName(rel.getTargetURI());
                       parts.add( rel.getPackage().getPart(relName) );
                    }
                 }
                 for(PackageRelationship rel : part.getRelationshipsByType(XSSFRelation.VML_DRAWINGS.getRelation())) {
                    if(rel.getTargetMode() == TargetMode.INTERNAL) {
                       PackagePartName relName = PackagingURIHelper.createPartName(rel.getTargetURI());
                       parts.add( rel.getPackage().getPart(relName) );
                    }
                 }
              } catch(InvalidFormatException e) {
                 throw new TikaException("Broken OOXML file", e);
              }
           }
       } catch (Exception e){

       }*/

       return parts;
    }

    @Override
    public MetadataExtractor getMetadataExtractor() {
        return new MetadataExtractor(extractor, TYPE) {
            @Override
            public void extract(Metadata metadata) throws TikaException {
                super.extract(metadata);
                //TODO apparantly the metadata Extractor is called before the sheets are parsed. Need to build an additional parser
                metadata.set(TikaMetadataKeys.PROTECTED, "false");
                if(getWorkbookProtection()){
                    metadata.set(TikaMetadataKeys.PROTECTED,"true");
                }
            }
        };
    }

    public boolean getWorkbookProtection(){
        return this.protectedWorkbook;
    }

   /**
    * Handler for sheets. Processes each row and cell,
    *  formatting Cells as best as it can.
    */
   public class EventBasedXSSFSheetHandler extends DefaultHandler {
       /**
        * Table with the styles used for formatting
        */
       private StylesTable stylesTable;

       private ReadOnlySharedStringsTable sharedStringsTable;

       // Set when V start element is seen
       private boolean vIsOpen;
       // Set when F start element is seen
       private boolean fIsOpen;

       // Set when cell start element is seen;
       // used when cell close element is seen.
       private int nextDataType;

       // Used to format numeric cell values.
       private short formatIndex;
       private String formatString;
//       private final DataFormatter formatter;

       // Gathers characters as they are seen.
       private StringBuffer value = new StringBuffer();
       private StringBuffer formula = new StringBuffer();

       private boolean firstCellOfRow;
       private boolean rowHasContent;
       private boolean protectedSheet;
        // Set when headerFooter element is seen
       private boolean hfIsOpen;

       private StringBuffer hfValue = new StringBuffer();
       private XHTMLContentHandler xhtml;


       /**
        * Accepts objects needed while parsing.
        *
        * @param styles  Table of styles
        * @param strings Table of shared strings
        */
       public EventBasedXSSFSheetHandler(
               StylesTable styles,
               ReadOnlySharedStringsTable strings,
               XHTMLContentHandler xhtml) {
           this.stylesTable = styles;
           this.sharedStringsTable = strings;
           this.nextDataType = Cell.CELL_TYPE_NUMERIC;
//           this.formatter = new DataFormatter();
           this.xhtml = xhtml;
           this.firstCellOfRow = false;
           this.rowHasContent = false;
       }

       public void startElement(String uri, String localName, String name,
                                Attributes attributes) throws SAXException {

           if ("inlineStr".equals(name) || "v".equals(name)) {
               vIsOpen = true;
               // Clear contents cache
               value.setLength(0);
           } else if ("f".equals(name)) {
              // Clear contents cache
              formula.setLength(0);

              // Mark us as being a formula if not already
              if(nextDataType == Cell.CELL_TYPE_NUMERIC) {
                 nextDataType = Cell.CELL_TYPE_FORMULA;
              }

              // Decide where to get the formula string from
              String type = attributes.getValue("t");
              if(type != null && type.equals("shared")) {
                 System.err.println("Warning - shared formulas not yet supported!");
              } else {
                 fIsOpen = true;
              }
           }
           else if("row".equals(name)) {
               firstCellOfRow = true;
               this.rowHasContent = false;
           }
           // c => cell
           else if ("c".equals(name)) {
               // Set up defaults.
               this.nextDataType = Cell.CELL_TYPE_NUMERIC;
               this.formatIndex = -1;
               this.formatString = null;

               String cellType = attributes.getValue("t");
               String cellStyleStr = attributes.getValue("s");
               if ("b".equals(cellType))
                   nextDataType = Cell.CELL_TYPE_BOOLEAN;
               else if ("e".equals(cellType))
                   nextDataType = Cell.CELL_TYPE_ERROR;
               else if ("inlineStr".equals(cellType))
                   nextDataType = Cell.CELL_TYPE_STRING;
                    //xssfDataType.INLINE_STRING;
               else if ("s".equals(cellType))
                   nextDataType = Cell.CELL_TYPE_STRING;
                   //xssfDataType.SST_STRING;
               else if ("str".equals(cellType))
                   nextDataType = Cell.CELL_TYPE_FORMULA;
                    //xssfDataType.FORMULA;
               else if (cellStyleStr != null) {
                  // Number, but almost certainly with a special style or format
                   int styleIndex = Integer.parseInt(cellStyleStr);
                   XSSFCellStyle style = stylesTable.getStyleAt(styleIndex);
                   this.formatIndex = style.getDataFormat();
                   this.formatString = style.getDataFormatString();
                   if (this.formatString == null)
                       this.formatString = BuiltinFormats.getBuiltinFormat(this.formatIndex);
               }
           } else if ("sheetProtection".equals(name)){
               hfValue.setLength(0);
               this.protectedSheet = true;
           } else if ("firstHeader".equals(name)){
               hfValue.setLength(0);
               hfIsOpen = true;
           } else if ("evenHeader".equals(name)){
               hfValue.setLength(0);
               hfIsOpen = true;
           } else if ("oddHeader".equals(name)){
               hfValue.setLength(0);
               hfIsOpen = true;
           } else if ("firstFooter".equals(name)){
               hfValue.setLength(0);
               hfIsOpen = true;
           } else if ("evenFooter".equals(name)){
               hfValue.setLength(0);
               hfIsOpen = true;
           } else if ("oddFooter".equals(name)){
               hfValue.setLength(0);
               hfIsOpen = true;
           }

       }

       public void endElement(String uri, String localName, String name)
               throws SAXException {
           String thisStr = null;

            // v => contents of a cell
           if ("v".equals(name)) {
               vIsOpen = false;
               // Process the value contents as required, now we have it all
               switch (nextDataType) {
                   case Cell.CELL_TYPE_BOOLEAN:
                       char first = value.charAt(0);
                       thisStr = first == '0' ? "FALSE" : "TRUE";
                       addCellToXHTML(thisStr);
                       break;

                   case Cell.CELL_TYPE_ERROR:
                       thisStr = "ERROR:" + value.toString();
                       addCellToXHTML(thisStr);
                       break;

                   case Cell.CELL_TYPE_FORMULA:
                       //TODO Checking how this works with SAX Parsing
                       // //if(formulasNotResults) {
                       //   thisStr = formula.toString();
                       //} else {
                          thisStr = value.toString();
                       //}
                       addCellToXHTML(thisStr);
                       break;

                     //TODO in org.apache.poi.ss.usermodel.Cell there is no INLINE_STRING type.
/*                   case INLINE_STRING:
                       cell.setCellType(xssfDataType.INLINE_STRING.toString());
                       
                       XSSFRichTextString rtsi = new XSSFRichTextString(value.toString());
                       thisStr = rtsi.toString();
                       break;*/

                   case Cell.CELL_TYPE_STRING:
                       String sstIndex = value.toString();
                       try {
                           int idx = Integer.parseInt(sstIndex);
                           XSSFRichTextString rtss = new XSSFRichTextString(sharedStringsTable.getEntryAt(idx));
                           thisStr = rtss.toString();
                           addCellToXHTML(thisStr);
                       }
                       catch (NumberFormatException ex) {
                           System.err.println("Failed to parse SST index '" + sstIndex + "': " + ex.toString());
                       }
                       break;

                   case Cell.CELL_TYPE_NUMERIC:
                       String n = value.toString();
                       if (this.formatString != null){
                           thisStr = formatter.formatRawCellContents(Double.parseDouble(n), this.formatIndex, this.formatString);
                       }
                       else {
                           thisStr = n;
                       }
                       addCellToXHTML(thisStr);
                       break;

                   default:
                       thisStr = "(TODO: Unexpected type: " + nextDataType + ")";
                       addCellToXHTML(thisStr);
                       break;
               }
           } else if ("f".equals(name)) {
              fIsOpen = false;
           } else if ("row".equals(name)) {
               if(this.rowHasContent){
                   xhtml.endElement("tr");
               }
           } else if ("firstHeader".equals(name)){
               hfIsOpen = false;
               xhtml.element("p", hfValue.toString());
           } else if ("evenHeader".equals(name)){
               hfIsOpen = false;
               xhtml.element("p", hfValue.toString());
           } else if ("oddHeader".equals(name)){
               hfIsOpen = false;
               xhtml.element("p", hfValue.toString());
           } else if ("firstFooter".equals(name)){
               hfIsOpen = false;
               xhtml.element("p", hfValue.toString());
           } else if ("evenFooter".equals(name)){
               hfIsOpen = false;
               xhtml.element("p", hfValue.toString());
           } else if ("oddFooter".equals(name)){
               hfIsOpen = false;
               xhtml.element("p", hfValue.toString());
           }
       }

       /**
        * Captures characters only if a suitable element is open.
        * Originally was just "v"; extended for inlineStr also.
        */
       public void characters(char[] ch, int start, int length)
               throws SAXException {
           if (vIsOpen) {
               value.append(ch, start, length);
           }
           if (fIsOpen) {
              formula.append(ch, start, length);
           }
           if (hfIsOpen) {
              hfValue.append(ch, start, length);
           }
       }

       public void addCellToXHTML(String cellContent){
            try{
               if(firstCellOfRow){
                   xhtml.startElement("tr");
                   this.firstCellOfRow=false;
               }
               xhtml.startElement("td");
               xhtml.characters(cellContent);
               xhtml.endElement("td");
               this.rowHasContent = true;
            } catch (SAXException e){

            }
       }

       public boolean isSheetProtected(){
           return this.protectedSheet;
       }
   }


}

